/*
 * Catalina - Simplified command line wrapper for the Catalina compilation 
 *            system for the Parallax Propeller. Turns catalina options
 *            into LCC options, and then invokes lcc.
 *
 * version 2.0 - initial release (to coincide with bind 2.0)
 *
 * version 2.1 - remove duplicate version number output on '-v'
 *
 * version 2.1 - fixed crash if an option that requires a value is not 
 *               followed by one (e.g. -D)
 *
 * version 2.2 - just update version number
 *
 * Version 2.3 - Allow for spaces in file/path names.
 *               Fix bug using incorrect LCCDIR for LARGE memory model.
 *               Allow -o to be used with -c and -S options.
 *               Allow upper or lowercase 'k' or 'm' to -M option.
 *
 * version 2.4 - Add '-g' option (passed to lcc, but enables listing). Also
 *               passed to binder to invoke debug file generator. Also sets
 *               the target.
 *               Add SMALL (and __CATALINA_SMALL) symbol when using SMALL
 *               memory model.
 *
 * version 2.4.1 - make it possible to override the target used when -g or
 *               -g3 is specified. If a different target has already been
 *               specified the -g options will not override it. Also, if a
 *               -t option is subsequently parsed, it takes precedence.
 *
 * Version 2.5 - just update version number.
 *
 * Version 2.6 - Add Optimizer option (-O).
 *
 * version 2.7 - just update version number.
 *
 * version 2.8 - intercept EEPROM and SDCARD and set layout (1 for EEPROM,
 *               6 for SDCARD).
 *
 * version 2.9 - Added -q option (to test the use of command line symbols) 
 *               and also defined the the PREPROCESS symbol to disable all 
 *               preprocessing code altogether (this code may be removed 
 *               altogether later).
 *
 *               Added statistics (segment and file sizes) generation, and
 *               the '-k' option to suppress them.
 *
 *               Added -R option, for setting the size to use for the 
 *               READ-WRITE segments - i.e. the CODE and CNST segments 
 *               will be shifted to begin at this address. This allows 
 *               memory layout 3 to be used with a combination of SPI 
 *               SRAM (Read/Write) and SPI FLASH (Read Only). For example, 
 *               the following command will put the read/write segments at 
 *               0x08000 and the read only segments at 0x10000:
 *
 *                  catalina hello_world.c -x3 -R32k
 *
 * Version 3.0   Add '-P' option to set base address of read-write segments.
 *               This is similar to the existing '-R' option which sets the
 *               base address of the read-only segments. Also, fixed some
 *               errors in calculating the segment sizes in the presence of
 *               the -P or -R command line options.
 *
 *               Added support for memory layout 4 (XMM SMALL). Now layout
 *               3 is for XMM LARGE mode and LAYOUT 4 is for XMM SMALL mode.
 *               Both these layouts are intended to be used for programs 
 *               that execute from using SPI FLASH, so both layout 3 and 
 *               layout 4 imply FLASH - so you don't need to specify FLASH 
 *               separately if you use -x3 or -x4.
 *
 *               However, if FLASH is specified on the command line, the 
 *               layout is set to 3 unless it has already been set to a 
 *               non-zero value. Also, the SDCARD and EEPROM options now 
 *               work the same way - i.e. they only set the layout if it 
 *               has not already been set to a non-zero value.
 *
 *               Also, the TINY, SMALL and LARGE memory model options can
 *               now be specified on the command line. These set the layout 
 *               to 0, 2 and 5 respectively unless the layout has already 
 *               been set to something incompatible with these. Also, if 
 *               the layout is 3 then SMALL will set it to 4, and if the 
 *               layout is 4 then LARGE will set it to 3. 
 *
 *               The effect of all these changes is that you can now use
 *               EITHER the layout option (-x) to specify the layout - OR - 
 *               use a combination of a memory model option (i.e. TINY, 
 *               SMALL or LARGE) and a loader option (i.e. EEPROM, SDCARD,
 *               FLASH). The relationship between the two ways of specifying
 *               the layout is as follows:
 *
 *               -x0    <==> -D TINY                 (this is the default)
 *               -x1    <==> -D TINY  -D EEPROM      (or just -D EEPROM)
 *               -x2    <==> -D SMALL
 *               -x3    <==> -D LARGE -D FLASH       (or just -D FLASH)
 *               -x4    <==> -D SMALL -D FLASH
 *               -x5    <==> -D LARGE 
 *               -x6    <==> -D TINY  -D SDCARD      (or just -D SDCARD)
 *              
 *               The -x option is basically now just a "shorthand" method.
 *               Note that mixing the two methods is NOT recommended. It
 *               may result in defining some command-line symbols twice.
 *               Also note that invalid combinations are NOT guaranteed to
 *               be detected in all cases!
 *
 * Version 3.0.4 Quote filenames containing spaces on unix as well as windows.
 *
 * Version 3.1 - just update version number.
 *
 * Version 3.2 - Make line size 4096 chars.
 *
 * Version 3.3 - just update version number.
 *
 * Version 3.4 - just update version number.
 *
 * Version 3.5 - Added -C for defining 'Catalina' symbols (which used to be 
 *               defined using -D). Made -D mean define 'C' symbols. Note that 
 *               catbind now also uses -C but that Homespun still use -D. 
 *
 * Version 3.6 - Fix up path processing to correctly override defaults in all
 *               cases (LARGE was not working), and redefine LIBRARY_PATH to
 *               mean the path to the libraries exlculding the final "\lib" 
 *               or "\large_lib" (this is determined by the memory mode).
 *
 * Version 3.7 - Added Compact Memory Model support (CMM).
 *               Allow optimization level 0 (to disable optiimization).
 *
 * Version 3.8 - Added Compact Memory Model support (CMM) to EEPROM and
 *               SDCARD loaders. Also added:
 *
 *               -x8    <==> -D COMPACT
 *               -x9    <==> -D COMPACT -D EEPROM
 *               -x10   <==> -D COMPACT -D SDCARD
 *
 * Version 3.9 - just update version number.
 *
 * Version 3.10 - just update version number.
 *
 * Version 3.11 - Add support for using openspin as assembler. This is now
 *                the default. 
 *                Remove bstc support.
 *                Remove compile-time support for file preprocessing - this 
 *                is no longer required since bstc support has been dropped,
 *                and both openspin and homespun support symbols defined on
 *                the command-line.
 *                Define the Catalina symols HOMESPUN__ or OPENSPIN__ (and
 *                the corresponding C symbols __CATALINA_HOMESPUN__ or 
 *                __CATALINA_OPENSPIN__) depending on whether the Homespun
 *                or Openspin assemblers are in use.
 *                Make "FLASH is incompatible with current layout" message
 *                a warning only, and ignore the symbol - this makes the 
 *                compilation of Catalyst utility programs much simpler,
 *                otherwise they would need special compilation options.
 *                Added more control over verbosity - default is now minimal
 *                output messages, -v means verbose, -v -v means very verbose.
 *
 * Version 3.12 - Renamed openspin to spinnaker.
 *
 */

/*--------------------------------------------------------------------------
    This file is part of Catalina.

    Copyright 2009 Ross Higson

    Catalina is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Catalina is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Catalina.  If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------------- */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define VERSION            "3.12"

#define MAX_LINELEN        4096

#ifdef WIN32_PATHS         /* define this on the command line for Windows */
#define DEFAULT_SEP        "\\"
#define DEFAULT_LCCDIR     "C:\\Program Files\\Catalina" // must match default used by LCC
#define MULT_PATH_SEP      ";"
#else
#define DEFAULT_SEP        "/"
#define DEFAULT_LCCDIR     "/usr/local/lib/catalina" // must match default used by LCC
#define MULT_PATH_SEP      ":"
#endif

#define LIB_SUFFIX         "lib"
#define TGT_SUFFIX         "target"
#define INC_SUFFIX         "include"
#define LARGE_LIB_SUFFIX   "large_lib"
#define COMPACT_LIB_SUFFIX "compact_lib"

#define DEBUG_TARGET       "blackcat"

#define DEFAULT_LCC_ENV    "LCCDIR" // can only display this - cannot set it as an option to LCC

#define DEFAULT_LIB_ENV    "CATALINA_LIBRARY"
#define DEFAULT_TGT_ENV    "CATALINA_TARGET"
#define DEFAULT_INC_ENV    "CATALINA_INCLUDE" 
#define DEFAULT_TMP_ENV    "CATALINA_TEMPDIR"
#define DEFAULT_OPT_ENV    "CATALINA_LCCOPT"
#define DEFAULT_DEF_ENV    "CATALINA_DEFINE"

/* global variables used when processing options - e.g. to warn about incompatible combinations */

static int verbose   = 0; // 1 => verbose output, 2 => very verbose output
static int diagnose  = 0; // 1 => diagnostic output

static int assembler = 1; // 0 => homespun, 1 => open spin (default)

static int comp_only = 0; // 0 => -c not specified, 1 => -c specified
static int asm_only  = 0; // 0 => -S not specified, 1 => -S specified

static int format    = 0; // 0 => default, 1 => binary, 2 => eeprom
static int layout    = 0; // 0 .. 6 for LMM, 8..10 for compact 
static int memory    = 0; // max size in bytes 
static int readonly  = 0; // offset of read-only segments
static int readwrite = 0; // offset of read-write segments
static int byte_swap = 0; // size of byte swap
static int listing   = 0; // generate listing
static int glevel    = 0; // debug level
static int olevel    = 0; // optimize level
static int suppress  = 0; // suppress statistics

static int flash = 0;   // -C FLASH appended
static int sdcard = 0;  // -C SDCARD appended
static int eeprom = 0;  // -C EEPROM appended

static int output_named    = 0; // have name for output
static int output_override = 0; // default output name overridden
static int target_named    = 0; // have name for target

static char tgt_name[MAX_LINELEN + 1] = "";
static char out_name[MAX_LINELEN + 1] = "";

static char tgt_path[MAX_LINELEN + 1] = "";
static char lib_path[MAX_LINELEN + 1] = "";
static char inc_path[MAX_LINELEN + 1] = "";
static char tmp_path[MAX_LINELEN + 1] = "";
static char lcc_path[MAX_LINELEN + 1] = "";

static char def_tgt_path[MAX_LINELEN + 1] = "";
static char def_lib_path[MAX_LINELEN + 1] = "";
static char def_inc_path[MAX_LINELEN + 1] = "";

static char lcc_cmd[MAX_LINELEN + 1] = "";

void help(char *my_name) {
   fprintf(stderr, "\nusage: %s [option | file] ...\n\n", my_name);
   fprintf(stderr, "options:  -? or -h   print this help (and exit)\n");
   fprintf(stderr, "          -ah        assemble using homespun\n");
   fprintf(stderr, "          -as        assemble using spinnaker (default)\n");
   fprintf(stderr, "          -b         generate a binary output file (this is the default)\n");
   fprintf(stderr, "          -c         compile only (do not bind)\n");
   fprintf(stderr, "          -B width   byte swap (used with -F, width = 2, 4, 8)\n");
   fprintf(stderr, "          -d         output diagnostic messages\n");
   fprintf(stderr, "          -C symbol  define a Catalina symbol (e.g. -C HYDRA)\n");
   fprintf(stderr, "          -D symbol  define a symbol (e.g. -D printf=tiny_printf)\n");
   fprintf(stderr, "          -e         generate an eeprom output file\n");
   fprintf(stderr, "          -F format  convert output to another format (e.g. -F Intel)\n");
   fprintf(stderr, "          -g[level]  generate debugging information (default level = 1)\n");
   fprintf(stderr, "          -I path    path to include files (default '%s')\n", def_inc_path);
   fprintf(stderr, "          -k         kill (suppress) statistics output\n");
   fprintf(stderr, "          -l lib     search library lib when binding\n");
   fprintf(stderr, "          -L path    path to libraries (default '%s')\n", def_lib_path);
   fprintf(stderr, "          -M size    maximum memory size (use with -x)\n");
   fprintf(stderr, "          -o name    name of output file (default is first file name)\n");
   fprintf(stderr, "          -O[level]  optimize code (default level = 1)\n");
   fprintf(stderr, "          -P addr    address for Read-Write segments\n");
   fprintf(stderr, "          -R addr    address for Read-Only segments\n");
   fprintf(stderr, "          -S         compile to assembly code (do not bind)\n");
   fprintf(stderr, "          -t name    name of dedicated target to use\n");
   fprintf(stderr, "          -T path    path to target files (default '%s')\n", def_tgt_path);
   fprintf(stderr, "          -U symbol  undefine symbol (e.g. -U DEFAULT)\n");
   fprintf(stderr, "          -v         verbose (output information messages)\n");
   fprintf(stderr, "          -v -v      very verbose (more information messages)\n");
   fprintf(stderr, "          -W option  option to pass directly to LCC\n");
   fprintf(stderr, "          -x layout  use specified memory layout (layout = 0 .. 6, 8 .. 10)\n");
   fprintf(stderr, "          -y         generate listing file\n");
}

// safecpy will never write more than size characters, 
// and is guaranteed to null terminate its result, so
// make sure the buffer passed is at least size + 1
char * safecpy(char *dst, const char *src, size_t size) {
   dst[size] = '\0';
   if (src) {
      return strncpy(dst, src, size - strlen(dst));
   }
   return NULL;
}

// safecat will never write more than size characters, 
// and is guaranteed to null terminate its resul, so
// make sure the buffer passed is at least size + 1
char * safecat(char *dst, const char *src, size_t size) {
   dst[size] = '\0';
   if (src) {
      return strncat(dst, src, size - strlen(dst));
   }
   return NULL;
}

// pathcat will check for paths with spaces and quote
// them if we are using Win32 style path names, otherwise
// it is much the same as two safecats
void pathcat(char *dst, const char *src, const char *sfx, size_t max) {
//#ifdef WIN32_PATHS
   if ((strchr(src, ' ') != NULL) && (src[0] != '\"')) {
      safecat(dst, "\"", max);
      safecat(dst, src, max);
      if (sfx != NULL) {
         safecat(dst, sfx, max);
      }
      safecat(dst, "\"", max);
   }
   else {
      safecat(dst, src, max);
      if (sfx != NULL) {
         safecat(dst, sfx, max);
      }
   }
//#else
//   safecat(dst, src, max);
//   if (sfx != NULL) {
//      safecat(dst, sfx, max);
//   }
//#endif   
}

/*
 * process symbols that have special meanings, and return TRUE 
 * if the symbol is to be passed onto the compiler. Also, return
 * a value of -1 in the "code" integer if we encounter an error
 */
int pass_symbol_to_compiler(char *symbol, int *code) {
   int pass = 1;

   if (strcmp (symbol, "EEPROM") == 0) {
      if ((layout == 0)||(layout == 1)) {
         if (verbose) {
            fprintf(stderr, "EEPROM implies -x1\n");
         }
         layout = 1;
      }
      else if ((layout == 8)||(layout == 9)) {
         if (verbose) {
            fprintf(stderr, "EEPROM implies -x9\n");
         }
         layout = 9;
      }
      else if ((layout == 2)||(layout==5)) {
         // we detect this later (in the target)
      }
      else {
         fprintf(stderr, "EEPROM is incompatible with current layout\n");
         *code = -1;
      }
   }
   else if (strcmp (symbol, "FLASH") == 0) {
      if ((layout==0)||(layout==3)||(layout==5)) {
         if (verbose) {
            fprintf(stderr, "FLASH implies -x3\n");
         }
         layout = 3;
      }
      else if ((layout==2)||(layout==4)) {
         if (verbose) {
            fprintf(stderr, "FLASH implies -x4\n");
         }
         layout = 4;
      }
      else {
         fprintf(stderr, "FLASH is incompatible with current layout - ignoring\n");
      }
   }
   else if (strcmp (symbol, "SDCARD") == 0) {
      if ((layout == 0)||(layout==6)) {
         if (verbose) {
            fprintf(stderr, "SDCARD implies -x6\n");
         }
         layout = 6;
      }
      else if ((layout == 8)||(layout == 10)) {
         if (verbose) {
            fprintf(stderr, "SDCARD implies -x10\n");
         }
         layout = 10;
      }
      else {
         fprintf(stderr, "SDCARD is incompatible with current layout\n");
         *code = -1;
      }
   }
   else if (strcmp (symbol, "TINY") == 0) {
      pass = 0; // don't pass this symbol yet - we do it later
      if ((layout == 2)||(layout == 5)) {
         if (verbose) {
            fprintf(stderr, "TINY implies -x0\n");
         }
         layout = 0;
      }
      else if ((layout == 3)||(layout==4)) {
         fprintf(stderr, "TINY is incompatible with current layout\n");
         *code = -1;
      }
   }
   else if (strcmp (symbol, "SMALL") == 0) {
      pass = 0; // don't pass this symbol yet - we do it later
      if ((layout == 0)||(layout==1)||(layout == 2)||(layout == 5)) {
         if (verbose) {
            fprintf(stderr, "SMALL implies -x2\n");
         }
         layout = 2;
      }
      else if ((layout == 3)||(layout == 4)) {
         if (verbose) {
            fprintf(stderr, "SMALL implies -x4\n");
         }
         layout = 4;
      }
      else {
         fprintf(stderr, "SMALL is incompatible with current layout\n");
         *code = -1;
      }
   }
   else if (strcmp (symbol, "LARGE") == 0) {
      pass = 0; // don't pass this symbol yet - we do it later
      if ((layout == 0)||(layout==1)||(layout == 2)||(layout==5)) {
         if (verbose) {
            fprintf(stderr, "LARGE implies -x5\n");
         }
         layout = 5;
      }
      else if ((layout == 3)||(layout == 4)) {
         if (verbose) {
            fprintf(stderr, "LARGE implies -x3\n");
         }
         layout = 3;
      }
      else {
         fprintf(stderr, "LARGE is incompatible with current layout\n");
         *code = -1;
      }
   }
   else if (strcmp (symbol, "COMPACT") == 0) {
      pass = 0; // don't pass this symbol yet - we do it later
      if (layout == 0) {
         if (verbose) {
            fprintf(stderr, "COMPACT implies -x8\n");
         }
         layout = 8;
      }
      else if (layout == 1) {
         if (verbose) {
            fprintf(stderr, "COMPACT implies -x9\n");
         }
         layout = 9;
      }
      else if (layout == 6) {
         if (verbose) {
            fprintf(stderr, "COMPACT implies -x10\n");
         }
         layout = 10;
      }
      else if ((layout != 8)&&(layout != 9)&&(layout != 10)) {
         fprintf(stderr, "COMPACT is incompatible with current layout\n");
         *code = -1;
      }
   }
   return pass;
   
}

/*
 * decode arguments, appending options to lcc command as we go -
 * return -1 if there is no further processing to do
 */
int decode_arguments (int argc, char *argv[]) {
   char   option[MAX_LINELEN + 3 + 1];
   char   optnum[3];
   char   temp_path[MAX_LINELEN + 1];
   int    code = 0;
   int    i = 0;
   int    len = 0;
   char   modifier;

   fprintf(stderr, "Catalina Compiler %s\n", VERSION); 

   if (argc == 1) {
      if (strlen(argv[0]) == 0) {
         // in case my name was not passed in ...
         help("catalina");
      }
      else {
         help(argv[0]);
      }
      code = -1;

   }
   while ((code >= 0) && (argc--)) {
      if (diagnose) {
         fprintf(stderr, "arg: %s\n", argv[i]);
      }
      if (i > 0) {
         if (argv[i][0] == '-') {
            if (diagnose) {
               fprintf(stderr, "switch: %s\n", argv[i]);
            }
            // it's a command line switch
            switch (argv[i][1]) {
               case 'h':
                  /* fall through to ... */

               case '?':
                  if (strlen(argv[0]) == 0) {
                     // in case my name was not passed in ...
                     help("catalina");
                  }
                  else {
                     help(argv[0]);
                  }
                  code = -1;
                  break;

               case 'a':
                  if (strlen(argv[i]) == 3) {
                     // use next char
                     if (argv[i][2] == 'h') {
                        assembler = 0;
                     }
                     else if (argv[i][2] == 's') {
                        assembler = 1;
                     }
                     else {
                        fprintf(stderr, "unrecognized assembler option - using spinnaker\n");
                        assembler = 1;
                     }
                  }
                  else {
                      fprintf(stderr, "unrecognized assembler option - using spinnaker\n");
                      assembler = 1;
                  }
                  break;

               case 'b':
                  format = 1;
                  if (verbose) {
                     fprintf(stderr, "binary output format selected\n");
                  }
                  break;

               case 'B':
                  if (strlen(argv[i]) == 2) {
                     // use next arg
                     if (argc > 0) {
                        sscanf(argv[++i], "%d", &byte_swap);
                        argc--;
                     }
                     else {
                        fprintf(stderr, "option -B requires an argument\n");
                        code = -1;
                        break;
                     }
                  }
                  else {
                     // use remainder of this arg
                     sscanf(&argv[i][2], "%d", &byte_swap);
                  }
                  if (verbose) {
                     fprintf(stderr, "byte swap width = %d\n", byte_swap);
                  }
                  break;

               case 'c':
                  comp_only = 1;
                  if (verbose) {
                     fprintf(stderr, "compile only\n");
                  }
                  safecat(lcc_cmd, "-c ", MAX_LINELEN);
                  break;

               case 'd':
                  diagnose++;   /* increase diagnosis level */
                  verbose = 1;   /* diagnose implies verbose */
                  fprintf(stderr, "diagnostic level %d\n", diagnose);
                  break;

               case 'C':
                  if (strlen(argv[i]) == 2) {
                     if (argc > 0) {
                        i++;
                        // use next arg
                        // Note that we define the symbol both for LCC and also for the 
                        // binder, but for LCC we prefix it with "__CATALINA_" to avoid
                        // colliding with user symbols
                        if (pass_symbol_to_compiler(argv[i], &code)) {
                           safecat(lcc_cmd, "-D__CATALINA_", MAX_LINELEN);
                           safecat(lcc_cmd, argv[i], MAX_LINELEN);
                           safecat(lcc_cmd, " -Wl-C", MAX_LINELEN);
                           safecat(lcc_cmd, argv[i], MAX_LINELEN);
                           safecat(lcc_cmd, " ", MAX_LINELEN);
                           if (verbose) {
                              fprintf(stderr, "define Catalina symbol %s\n", argv[i]);
                           }
                        }
                        argc--;
                     }
                     else {
                        fprintf(stderr, "option -C requires an argument\n");
                        code = -1;
                        break;
                     }
                  }
                  else {
                     // use remainder of this arg
                     // Note that we define the symbol both for LCC and also for the 
                     // binder, but for LCC we prefix it with "__CATALINA_" to avoid
                     // colliding with user symbols
                     if (pass_symbol_to_compiler(&argv[i][2], &code)) {
                        safecat(lcc_cmd, "-D__CATALINA_", MAX_LINELEN);
                        safecat(lcc_cmd, &argv[i][2], MAX_LINELEN);
                        safecat(lcc_cmd, " -Wl-C", MAX_LINELEN);
                        safecat(lcc_cmd, &argv[i][2], MAX_LINELEN);
                        safecat(lcc_cmd, " ", MAX_LINELEN);
                        if (verbose) {
                           fprintf(stderr, "define Catalina symbol %s\n", &argv[i][2]);
                        }
                     }
                  }
                  break;

               case 'D':
                  if (strlen(argv[i]) == 2) {
                     if (argc > 0) {
                        i++;
                        // use next arg
                           safecat(lcc_cmd, " -D", MAX_LINELEN);
                           safecat(lcc_cmd, argv[i], MAX_LINELEN);
                           safecat(lcc_cmd, " ", MAX_LINELEN);
                           if (verbose) {
                              fprintf(stderr, "define symbol %s\n", argv[i]);
                           }
                        argc--;
                     }
                     else {
                        fprintf(stderr, "option -D requires an argument\n");
                        code = -1;
                        break;
                     }
                  }
                  else {
                     // use remainder of this arg
                        safecat(lcc_cmd, " -D", MAX_LINELEN);
                        safecat(lcc_cmd, &argv[i][2], MAX_LINELEN);
                        safecat(lcc_cmd, " ", MAX_LINELEN);
                        if (verbose) {
                           fprintf(stderr, "define symbol %s\n", &argv[i][2]);
                        }
                  }
                  break;

               case 'e':
                  format = 2;
                  if (verbose) {
                     fprintf(stderr, "eeprom output format selected\n");
                  }
                  break;

               case 'F':
                  if (strlen(argv[i]) == 2) {
                     if (argc > 0) {
                        // use next arg
                        safecat(lcc_cmd, "-Wl-F", MAX_LINELEN);
                        safecat(lcc_cmd, argv[++i], MAX_LINELEN);
                        safecat(lcc_cmd, " ", MAX_LINELEN);
                        if (verbose) {
                           fprintf(stderr, "output format = %s\n", argv[i]);
                        }
                        argc--;
                     }
                     else {
                        fprintf(stderr, "option -F requires an argument\n");
                        code = -1;
                        break;
                     }
                  }
                  else {
                     // use remainder of this arg
                     safecat(lcc_cmd, "-Wl", MAX_LINELEN);
                     safecat(lcc_cmd, argv[i], MAX_LINELEN);
                     safecat(lcc_cmd, " ", MAX_LINELEN);
                     if (verbose) {
                        fprintf(stderr, "output format = %s\n", &argv[i][2]);
                     }
                  }
                  break;

               case 'I':
                  if (strlen(inc_path) != 0) {
                     safecat(inc_path, MULT_PATH_SEP, MAX_LINELEN);
                  }
                  if (strlen(argv[i]) == 2) {
                     if (argc > 0) {
                        // use next arg
                        safecat(inc_path, argv[++i], MAX_LINELEN);
                        if (verbose) {
                           fprintf(stderr, "adding include path %s\n", argv[i]);
                        }
                        argc--;
                     }
                     else {
                        fprintf(stderr, "option -I requires an argument\n");
                        code = -1;
                        break;
                     }
                  }
                  else {
                     // use remainder of this arg
                     safecat(inc_path, &argv[i][2], MAX_LINELEN);
                     if (verbose) {
                        fprintf(stderr, "adding include path %s\n", &argv[i][2]);
                     }
                  }
                  break;

               case 'k':
                  suppress = 1;
                  if (verbose) {
                     fprintf(stderr, "suppress statistics\n");
                  }
                  break;

               case 'l':
                  if (strlen(argv[i]) == 2) {
                     if (argc > 0) {
                        // use next arg
                        safecat(lcc_cmd, "-l", MAX_LINELEN);
                        safecat(lcc_cmd, argv[++i], MAX_LINELEN);
                        safecat(lcc_cmd, " ", MAX_LINELEN);
                        // Now define a library symbol for LCC and the binder
                        safecat(lcc_cmd, "-D__CATALINA_lib", MAX_LINELEN);
                        safecat(lcc_cmd, argv[i], MAX_LINELEN);
                        safecat(lcc_cmd, " -Wl-Clib", MAX_LINELEN);
                        safecat(lcc_cmd, argv[i], MAX_LINELEN);
                        safecat(lcc_cmd, " ", MAX_LINELEN);
                        if (verbose) {
                           fprintf(stderr, "search library lib%s\n", argv[i]);
                        }
                        argc--;
                     }
                     else {
                        fprintf(stderr, "option -l requires an argument\n");
                        code = -1;
                        break;
                     }
                  }
                  else {
                     // use remainder of this arg
                     safecat(lcc_cmd, argv[i], MAX_LINELEN);
                     safecat(lcc_cmd, " ", MAX_LINELEN);
                     // Now define a library symbol for LCC and the binder
                     safecat(lcc_cmd, "-D__CATALINA_lib", MAX_LINELEN);
                     safecat(lcc_cmd, &argv[i][2], MAX_LINELEN);
                     safecat(lcc_cmd, " -Wl-Clib", MAX_LINELEN);
                     safecat(lcc_cmd, &argv[i][2], MAX_LINELEN);
                     safecat(lcc_cmd, " ", MAX_LINELEN);
                     if (verbose) {
                        fprintf(stderr, "search library lib%s\n", &argv[i][2]);
                     }
                  }
                  break;

               case 'L':
                  if (strlen(argv[i]) == 2) {
                     if (argc > 0) {
                        // use next arg
                        safecpy(lib_path, argv[++i], MAX_LINELEN);
                        argc--;
                     }
                     else {
                        fprintf(stderr, "option -L requires an argument\n");
                        code = -1;
                        break;
                     }
                  }
                  else {
                     // use remainder of this arg
                     safecpy(lib_path, &argv[i][2], MAX_LINELEN);
                  }
                  if (verbose) {
                     fprintf(stderr, "setting library path = %s\n", lib_path);
                  }
                  break;

               case 'M':
                  if (strlen(argv[i]) == 2) {
                     if (argc > 0) {
                        // use next arg
                        sscanf(argv[++i], "%d%c", &memory, &modifier);
                        argc--;
                     }
                     else {
                        fprintf(stderr, "option -M requires an argument\n");
                        code = -1;
                        break;
                     }
                  }
                  else {
                     // use remainder of this arg
                     sscanf(&argv[i][2], "%d%c", &memory, &modifier);
                  }
                  if (tolower(modifier) == 'k') {
                     memory *= 1024;
                  }
                  else if (tolower(modifier) == 'm') {
                     memory *= 1024 * 1024;
                  }
                  if (verbose) {
                     fprintf(stderr, "maximum memory = %d\n", memory);
                  }
                  break;

               case 'P':
                  if (strlen(argv[i]) == 2) {
                     // use next arg
                     if (argc > 0) {
                        sscanf(argv[++i], "%d%c", &readwrite, &modifier);
                     }
                     else {
                        fprintf(stderr, "option -P requires an argument\n");
                        code = -1;
                        break;
                     }
                     argc--;
                  }
                  else {
                     // use remainder of this arg
                     sscanf(&argv[i][2], "%d%c", &readwrite, &modifier);
                  }
                  if (tolower(modifier) == 'k') {
                     readwrite *= 1024;
                  }
                  else if (tolower(modifier) == 'm') {
                     readwrite *= 1024 * 1024;
                  }
                  if (verbose) {
                     fprintf(stderr, "readwrite address = %d\n", readwrite);
                  }
                  break;

               case 'R':
                  if (strlen(argv[i]) == 2) {
                     // use next arg
                     if (argc > 0) {
                        sscanf(argv[++i], "%d%c", &readonly, &modifier);
                     }
                     else {
                        fprintf(stderr, "option -R requires an argument\n");
                        code = -1;
                        break;
                     }
                     argc--;
                  }
                  else {
                     // use remainder of this arg
                     sscanf(&argv[i][2], "%d%c", &readonly, &modifier);
                  }
                  if (tolower(modifier) == 'k') {
                     readonly *= 1024;
                  }
                  else if (tolower(modifier) == 'm') {
                     readonly *= 1024 * 1024;
                  }
                  if (verbose) {
                     fprintf(stderr, "readonly address = %d\n", readonly);
                  }
                  break;

               case 'o':
                  output_named = 1;
                  output_override = 1;
                  if (strlen(argv[i]) == 2) {
                     if (argc > 0) {
                        // use next arg
                        safecpy(out_name,"", MAX_LINELEN);
                        pathcat(out_name, argv[++i], NULL, MAX_LINELEN);
                        argc--;
                     }
                     else {
                        fprintf(stderr, "option -o requires an argument\n");
                        code = -1;
                        break;
                     }
                  }
                  else {
                     // use remainder of this arg
                     safecpy(out_name,"", MAX_LINELEN);
                     pathcat(out_name, &argv[i][2], NULL, MAX_LINELEN);
                  }
                  if (verbose) {
                     fprintf(stderr, "output file = %s\n", out_name);
                  }
                  break;

               case 'O':
                  if (verbose) {
                     fprintf(stderr, "optimize code - listing selected\n");
                  }
                  listing = 1;
                  if (strlen(argv[i]) == 2) {
                     olevel = 1;
                  }
                  else {
                     // use remainder of this arg
                     sscanf(&argv[i][2], "%d", &olevel);
                  }
                  if ((olevel < 0) || (olevel > 9)) {
                     olevel = 1; // olevel must be 0 to 9
                  }
                  optnum[0] = '0' + olevel;
                  optnum[1] = ' ';
                  optnum[2] = '\0';
                  safecat(lcc_cmd, " -Wl-O", MAX_LINELEN);
                  safecat(lcc_cmd, optnum, MAX_LINELEN);
                  break;

               case 'S':
                  asm_only = 1;
                  if (verbose) {
                     fprintf(stderr, "compile to assembly only\n");
                  }
                  safecat(lcc_cmd, "-S ", MAX_LINELEN);
                  break;

               case 't':
                  if (target_named) {
                     fprintf(stderr, "option -t will override current target (%s)\n", tgt_name);
                  }
                  target_named = 1;
                  if (strlen(argv[i]) == 2) {
                     if (argc > 0) {
                        // use next arg
                        safecpy(tgt_name, argv[++i], MAX_LINELEN);
                        argc--;
                     }
                     else {
                        fprintf(stderr, "option -t requires an argument\n");
                        code = -1;
                        break;
                     }
                  }
                  else {
                     // use remainder of this arg
                     safecpy(tgt_name, &argv[i][2], MAX_LINELEN);
                  }
                  if (verbose) {
                     fprintf(stderr, "target name = %s\n", tgt_name);
                  }
                  break;

               case 'T':
                  if (strlen(argv[i]) == 2) {
                     if (argc > 0) {
                        // use next arg
                        safecpy(tgt_path, argv[++i], MAX_LINELEN);
                        argc--;
                     }
                     else {
                        fprintf(stderr, "option -T require an argument\n");
                        code = -1;
                        break;
                     }
                  }
                  else {
                     // use remainder of this arg
                     safecpy(tgt_path, &argv[i][2], MAX_LINELEN);
                  }
                  if (verbose) {
                     fprintf(stderr, "setting target path = %s\n", tgt_path);
                  }
                  break;

               case 'U':
                  if (strlen(argv[i]) == 2) {
                     if (argc > 0) {
                        // use next arg
                        // Note that we undefine the symbol both for LCC and also for the 
                        // binder, but for LCC we prefix it with "__CATALINA_" to avoid
                        // colliding with user symbols
                        safecat(lcc_cmd, "-U__CATALINA_", MAX_LINELEN);
                        safecat(lcc_cmd, argv[++i], MAX_LINELEN);
                        safecat(lcc_cmd, " -Wl-U", MAX_LINELEN);
                        safecat(lcc_cmd, argv[i], MAX_LINELEN);
                        safecat(lcc_cmd, " ", MAX_LINELEN);
                        if (verbose) {
                           fprintf(stderr, "undefine symbol %s\n", argv[i]);
                        }
                        argc--;
                     }
                     else {
                        fprintf(stderr, "option -U requires an argument\n");
                        code = -1;
                        break;
                     }
                  }
                  else {
                     // use remainder of this arg
                     // Note that we undefine the symbol both for LCC and also for the 
                     // binder, but for LCC we prefix it with "__CATALINA_" to avoid
                     // colliding with user symbols
                     safecat(lcc_cmd, "-U__CATALINA_", MAX_LINELEN);
                     safecat(lcc_cmd, &argv[i][2], MAX_LINELEN);
                     safecat(lcc_cmd, " -Wl-U", MAX_LINELEN);
                     safecat(lcc_cmd, &argv[i][2], MAX_LINELEN);
                     safecat(lcc_cmd, " ", MAX_LINELEN);
                     if (verbose) {
                        fprintf(stderr, "undefine symbol %s\n", &argv[i][2]);
                     }
                  }
                  break;

               case 'v':
                  verbose++;
                  if (verbose == 1) {
                     fprintf(stderr, "verbose mode\n");
                  }
                  else {
                     fprintf(stderr, "very verbose mode\n");
                  }
                  break;

               case 'W':
                  if (strlen(argv[i]) == 2) {
                     if (argc > 0) {
                        // use next arg
                        safecat(lcc_cmd, argv[++i], MAX_LINELEN);
                        safecat(lcc_cmd, " ", MAX_LINELEN);
                        if (verbose) {
                           fprintf(stderr, "passing option %s\n", argv[i]);
                        }
                        argc--;
                     }
                     else {
                        fprintf(stderr, "option -W requires an argument\n");
                        code = -1;
                        break;
                     }
                  }
                  else {
                     // use remainder of this arg
                     safecat(lcc_cmd, &argv[i][2], MAX_LINELEN);
                     safecat(lcc_cmd, " ", MAX_LINELEN);
                     if (verbose) {
                        fprintf(stderr, "passing option %s\n", &argv[i][2]);
                     }
                  }
                  break;

               case 'x':
                  if (strlen(argv[i]) == 2) {
                     if (argc > 0) {
                        // use next arg
                        sscanf(argv[++i], "%d", &layout);
                        argc--;
                     }
                     else {
                        fprintf(stderr, "option -x requires an argument\n");
                        code = -1;
                        break;
                     }
                  }
                  else {
                     // use remainder of this arg
                     sscanf(&argv[i][2], "%d", &layout);
                  }
                  if (verbose) {
                     fprintf(stderr, "memory layout %d\n", layout);
                  }
                  if ((layout < 0) || (layout == 7) || (layout > 10)) {
                     fprintf(stderr, "unknown memory layout - using layout 0\n");
                     layout = 0;
                  }
                  if (layout == 1) {
                     if (!eeprom) {
                        if (verbose) {
                           fprintf(stderr, "option -x1 implies -C EEPROM\n");
                        }
                        safecat(lcc_cmd, "-D__CATALINA_EEPROM -Wl-CEEPROM ", MAX_LINELEN);
                        eeprom = 1;
                     }
                  }
                  if (layout == 3) {
                     if (!flash) {
                        if (verbose) {
                           fprintf(stderr, "option -x3 implies -C FLASH\n");
                        }
                        safecat(lcc_cmd, "-D__CATALINA_FLASH -Wl-CFLASH ", MAX_LINELEN);
                        flash = 1;
                     }
                  }
                  if (layout == 4) {
                     if (!flash) {
                        if (verbose) {
                           fprintf(stderr, "option -x4 implies -C FLASH\n");
                        }
                        safecat(lcc_cmd, "-D__CATALINA_FLASH -Wl-CFLASH ", MAX_LINELEN);
                        flash = 1;
                     }
                  }
                  if (layout == 6) {
                     if (!sdcard) {
                        if (verbose) {
                           fprintf(stderr, "option -x6 implies -C SDCARD\n");
                        }
                        safecat(lcc_cmd, "-D__CATALINA_SDCARD -Wl-CSDCARD ", MAX_LINELEN);
                        sdcard = 1;
                     }
                  }
                  if (layout == 9) {
                     if (!eeprom) {
                        if (verbose) {
                           fprintf(stderr, "option -x9 implies -C EEPROM\n");
                        }
                        safecat(lcc_cmd, "-D__CATALINA_EEPROM -Wl-CEEPROM ", MAX_LINELEN);
                        eeprom = 1;
                     }
                  }
                  if (layout == 10) {
                     if (!sdcard) {
                        if (verbose) {
                           fprintf(stderr, "option -x10 implies -C SDCARD\n");
                        }
                        safecat(lcc_cmd, "-D__CATALINA_SDCARD -Wl-CSDCARD ", MAX_LINELEN);
                        sdcard = 1;
                     }
                  }
                  break;

               case 'g':
                  if (verbose) {
                     fprintf(stderr, "generate debug information - listing selected\n");
                  }
                  listing = 1;
                  if (strlen(argv[i]) == 2) {
                     glevel = 1;
                  }
                  else {
                     // use remainder of this arg
                     sscanf(&argv[i][2], "%d", &glevel);
                  }
                  if ((glevel <= 0) || (glevel > 9)) {
                     glevel = 1; // glevel must be 1 to 9
                  }
                  optnum[0] = '0' + glevel;
                  optnum[1] = ' ';
                  optnum[2] = '\0';
                  safecat(lcc_cmd, argv[i], MAX_LINELEN);
                  safecat(lcc_cmd, " -Wl-g", MAX_LINELEN);
                  safecat(lcc_cmd, optnum, MAX_LINELEN);
                  safecat(lcc_cmd, " -Wf-g", MAX_LINELEN);
                  safecat(lcc_cmd, optnum, MAX_LINELEN);
                  safecat(lcc_cmd, " -D__CATALINA_BLACKBOX -Wl-CBLACKBOX ", MAX_LINELEN);
                  if (target_named) {
                     fprintf(stderr, "option -g will NOT override current target (%s)\n", tgt_name);
                  }
                  else {
                     safecpy(tgt_name, DEBUG_TARGET, MAX_LINELEN);
                     target_named = 1;
                  }
                  break;

               case 'y':
                  listing = 1;
                  if (verbose) {
                     fprintf(stderr, "listing selected\n");
                  }
                  break;

               default:
                  // unrecognized switch - just pass it on
                  if (verbose) {
                     fprintf(stderr, "passing switch %s\n", argv[i]);
                  }
                  pathcat(lcc_cmd, argv[i], NULL, MAX_LINELEN);
                  safecat(lcc_cmd, " ", MAX_LINELEN);
                  break;

            }
         }
         else {
            // assume its a filename
            if (output_named == 0) {
               output_named = 1;
               safecpy(out_name, (argv[i]), MAX_LINELEN);
               len = strlen(out_name);
               if ((len > 2) && (out_name[len-2] == '.') && (out_name[len-1] = 'c')) {
                  out_name[len-2] = '\0';
                  if (verbose) {
                     fprintf(stderr, "setting output to %s\n", out_name);
                  }
               }
            }
            pathcat(lcc_cmd, argv[i], NULL, MAX_LINELEN);
            safecat(lcc_cmd, " ", MAX_LINELEN);
            if (verbose) {
               fprintf(stderr, "passing filename %s\n", argv[i]);
            }
         }
      }
      i++; // next argument
   }

   // now finalize accumulated options 

   // append preset paths
   temp_path[0]='\0';
   safecpy(temp_path, getenv(DEFAULT_INC_ENV), MAX_LINELEN);
   if (strlen(temp_path) != 0) {
      if (strlen(inc_path) != 0) {
         safecat(inc_path, MULT_PATH_SEP, MAX_LINELEN);
      }
      safecat(inc_path, temp_path, MAX_LINELEN);
   }

   if (verbose == 1) {
      safecat(lcc_cmd, "-Wl-v ", MAX_LINELEN);
   }
   else if (verbose > 1) {
      safecat(lcc_cmd, "-Wl-v -Wl-v ", MAX_LINELEN);
   }
   if (suppress == 1) {
      safecat(lcc_cmd, "-Wl-k ", MAX_LINELEN);
   }
   if (diagnose > 0) {
      safecat(lcc_cmd, "-v ", MAX_LINELEN);
      for (i = 0; i < diagnose; i++) {
         safecat(lcc_cmd, "-Wl-d ", MAX_LINELEN);
      }
   }
   if (assembler == 0) {
      safecat(lcc_cmd, "-D__CATALINA_HOMESPUN__ -Wl-CHOMESPUN__ -Wl-ah ", MAX_LINELEN);
   }
   if (assembler == 1) {
      safecat(lcc_cmd, "-D__CATALINA_OPENSPIN__ -Wl-COPENSPIN__ -Wl-as ", MAX_LINELEN);
   }
   if (format == 1) {
      safecat(lcc_cmd, "-Wl-w-b ", MAX_LINELEN);
   }
   if (format == 2){
      safecat(lcc_cmd, "-Wl-w-e ", MAX_LINELEN);
   }
   if (memory > 0) {
      sprintf(option, "-Wl-M%d ", memory);
      safecat(lcc_cmd, option, MAX_LINELEN);
   }
   if (readonly > 0) {
      sprintf(option, "-Wl-R%d ", readonly);
      safecat(lcc_cmd, option, MAX_LINELEN);
   }
   if (readwrite > 0) {
      sprintf(option, "-Wl-P%d ", readwrite);
      safecat(lcc_cmd, option, MAX_LINELEN);
   }
   if (byte_swap > 0) {
      sprintf(option, "-Wl-B%d ", byte_swap);
      safecat(lcc_cmd, option, MAX_LINELEN);
   }
   if (listing == 1) {
      if (assembler == 0) {
         safecat(lcc_cmd, "-Wl-w-d ", MAX_LINELEN);
      }
      if (assembler == 1) {
         safecat(lcc_cmd, "-Wl-w-l ", MAX_LINELEN);
      }
   }
   if (diagnose) {
      fprintf(stderr, "executable name = %s\n", argv[0]);
   }
   if (output_named) {
      if ((output_override) || ((comp_only == 0) && (asm_only == 0))) {
         safecat(lcc_cmd, "-o ", MAX_LINELEN);
         pathcat(lcc_cmd, out_name, NULL, MAX_LINELEN);
         safecat(lcc_cmd, " ", MAX_LINELEN);
      }
   }
   if (target_named) {
      safecat(lcc_cmd, "-Wl-t", MAX_LINELEN);
      pathcat(lcc_cmd, tgt_name, NULL, MAX_LINELEN);
      safecat(lcc_cmd, " ", MAX_LINELEN);
   }
   return code;
}


int main (int argc, char *argv[]) {
   int  result = 0;
   char temp_env[MAX_LINELEN + 1] = "";
   char option[MAX_LINELEN + 3 + 1];
   char path_sfx[MAX_LINELEN + 1];
   char *define;

   // get default path values from the environment
   safecpy(lcc_path, getenv(DEFAULT_LCC_ENV), MAX_LINELEN);
   safecpy(lib_path, getenv(DEFAULT_LIB_ENV), MAX_LINELEN);
   safecpy(tmp_path, getenv(DEFAULT_TMP_ENV), MAX_LINELEN);
   safecpy(tgt_path, getenv(DEFAULT_TGT_ENV), MAX_LINELEN);

   // set the default LCC path if not set from the environment
   if (strlen(lcc_path) == 0) {
      safecpy(lcc_path, DEFAULT_LCCDIR, MAX_LINELEN);
   }

   // set up the LCC command with any default options from the environment
   safecpy(lcc_cmd, "lcc ", MAX_LINELEN);
   safecat(lcc_cmd, getenv(DEFAULT_OPT_ENV), MAX_LINELEN);
   safecat(lcc_cmd, " ", MAX_LINELEN);

   // set up the default path (also used to display the help defaults)
   safecpy(def_tgt_path, lcc_path, MAX_LINELEN);
   safecat(def_tgt_path, DEFAULT_SEP, MAX_LINELEN);
   safecat(def_tgt_path, TGT_SUFFIX, MAX_LINELEN);
   safecpy(def_lib_path, lcc_path, MAX_LINELEN);
   safecpy(def_inc_path, lcc_path, MAX_LINELEN);
   safecat(def_inc_path, DEFAULT_SEP, MAX_LINELEN);
   safecat(def_inc_path, INC_SUFFIX, MAX_LINELEN);

   // decode the arguments (which may override the default paths)
   result = decode_arguments(argc, argv);
   if (diagnose) {
      fprintf(stderr, "result of decoding arguments = %d\n", result);
   }

   // if the paths have not been set from the environment 
   // or on the command line, then use the defaults
   if (strlen(lib_path) == 0) {
      safecpy(lib_path, def_lib_path, MAX_LINELEN);
   }
   if (strlen(tgt_path) == 0) {
      safecpy(tgt_path, def_tgt_path, MAX_LINELEN);
   }
   if (strlen(inc_path) == 0) {
      safecpy(inc_path, def_inc_path, MAX_LINELEN);
   }

   // add the paths to the LCC command
   if (strlen(tgt_path) > 0) {
      safecat(lcc_cmd, "-Wl-T", MAX_LINELEN);
      pathcat(lcc_cmd, tgt_path, NULL, MAX_LINELEN);
      safecat(lcc_cmd, " ", MAX_LINELEN);
   }
   if (strlen(tmp_path) > 0) {
      safecat(lcc_cmd, "-tempdir=", MAX_LINELEN);
      pathcat(lcc_cmd, tmp_path, NULL, MAX_LINELEN);
      safecat(lcc_cmd, " ", MAX_LINELEN);
   }
   if (strlen(inc_path) > 0) {
      safecat(lcc_cmd, "-I", MAX_LINELEN);
      pathcat(lcc_cmd, inc_path, NULL, MAX_LINELEN);
      safecat(lcc_cmd, " ", MAX_LINELEN);
   }

   // get any Spin symbols defined by the environment variables
   safecpy(temp_env, getenv(DEFAULT_DEF_ENV), MAX_LINELEN);
   define = strtok(temp_env, " :;,");
   while (define != NULL) {
      // Note that we define the Spin symbol both for LCC and also for the
      // binder, but for LCC we preceed the symbol with "__CATALINA_"
      // to avoid colliding with any user symbols
      if (pass_symbol_to_compiler(define, &result)) {
         safecat(lcc_cmd, "-D__CATALINA_", MAX_LINELEN);
         safecat(lcc_cmd, define, MAX_LINELEN);
         safecat(lcc_cmd, " -Wl-C", MAX_LINELEN);
         safecat(lcc_cmd, define, MAX_LINELEN);
         safecat(lcc_cmd, " ", MAX_LINELEN);
         if (verbose) {
            fprintf(stderr, "define symbol %s\n", define);
         }
      }
      define = strtok(NULL, " :;,");
   }
   if (diagnose) {
      fprintf(stderr, "result of decoding %s = %d\n", DEFAULT_DEF_ENV, result);
   }
   if (result < 0) {
      if (diagnose) {
         fprintf(stderr, "%s exiting\n", argv[0]);
      }
      exit(result);
   }

   // specify layout and library path
   sprintf(option, "-Wl-x%d ", layout);
   safecat(lcc_cmd, option, MAX_LINELEN);
   if ((layout == 0) || (layout == 1) || (layout == 6)) {
      // use normal model code generator, and indicate LMM (TINY)
      safecat(lcc_cmd, "-Wl-CTINY -D__CATALINA_TINY ", MAX_LINELEN);
      if ((asm_only == 0) && (comp_only == 0)) {
         safecat(lcc_cmd, "-L", MAX_LINELEN);
         pathcat(lcc_cmd, lib_path, NULL, MAX_LINELEN);
         safecat(lcc_cmd, DEFAULT_SEP, MAX_LINELEN);
         safecat(lcc_cmd, LIB_SUFFIX, MAX_LINELEN);
         safecat(lcc_cmd, " ", MAX_LINELEN);
      }
   }
   else if ((layout == 2) || (layout == 4)) {
      // use normal model code generator, but indicate XMM (SMALL)
      safecat(lcc_cmd, "-Wl-CSMALL -D__CATALINA_SMALL ", MAX_LINELEN);
      if ((asm_only == 0) && (comp_only == 0)) {
         safecat(lcc_cmd, "-L", MAX_LINELEN);
         pathcat(lcc_cmd, lib_path, NULL, MAX_LINELEN);
         safecat(lcc_cmd, DEFAULT_SEP, MAX_LINELEN);
         safecat(lcc_cmd, LIB_SUFFIX, MAX_LINELEN);
         safecat(lcc_cmd, " ", MAX_LINELEN);
      }
   }
   else if ((layout == 3) || (layout == 5)) {
      // use large model code generator, and indicate XMM (LARGE)
      safecat(lcc_cmd, "-target=catalina_large/win32 -Wl-CLARGE -D__CATALINA_LARGE ", MAX_LINELEN);
      if ((asm_only == 0) && (comp_only == 0)) {
         safecat(lcc_cmd, "-L", MAX_LINELEN);
         pathcat(lcc_cmd, lib_path, NULL, MAX_LINELEN);
         safecat(lcc_cmd, DEFAULT_SEP, MAX_LINELEN);
         safecat(lcc_cmd, LARGE_LIB_SUFFIX, MAX_LINELEN);
         safecat(lcc_cmd, " ", MAX_LINELEN);
      }
   }
   else if ((layout == 8) || (layout == 9) || (layout == 10)) {
      // use compact code generator, and indicate CMM (COMPACT)
      safecat(lcc_cmd, "-target=catalina_compact/win32 -Wl-CCOMPACT -D__CATALINA_COMPACT ", MAX_LINELEN);
      if ((asm_only == 0) && (comp_only == 0)) {
         safecat(lcc_cmd, "-L", MAX_LINELEN);
         pathcat(lcc_cmd, lib_path, NULL, MAX_LINELEN);
         safecat(lcc_cmd, DEFAULT_SEP, MAX_LINELEN);
         safecat(lcc_cmd, COMPACT_LIB_SUFFIX, MAX_LINELEN);
         safecat(lcc_cmd, " ", MAX_LINELEN);
      }
   }
   else {
      fprintf(stderr, "Unknown memory layout %d\n", layout);
      exit(-1);
   }

   safecpy(temp_env, getenv(DEFAULT_INC_ENV), MAX_LINELEN);
   if (diagnose && strlen(temp_env) > 0) {
      fprintf(stderr, "%s = %s\n", DEFAULT_INC_ENV, temp_env);
   }
   temp_env[0]='\0';
   safecpy(temp_env, getenv(DEFAULT_LIB_ENV), MAX_LINELEN);
   if (diagnose && strlen(temp_env) > 0) {
      fprintf(stderr, "%s = %s\n", DEFAULT_LIB_ENV, temp_env);
   }
   temp_env[0]='\0';
   safecpy(temp_env, getenv(DEFAULT_TMP_ENV), MAX_LINELEN);
   if (diagnose && (strlen(temp_env) > 0)) {
      fprintf(stderr, "%s = %s\n", DEFAULT_TMP_ENV, temp_env);
   }
   temp_env[0]='\0';
   safecpy(temp_env, getenv(DEFAULT_TGT_ENV), MAX_LINELEN);
   if (diagnose && (strlen(temp_env) > 0)) {
      fprintf(stderr, "%s = %s\n", DEFAULT_TGT_ENV, temp_env);
   }
   temp_env[0]='\0';
   safecpy(temp_env, getenv(DEFAULT_LCC_ENV), MAX_LINELEN);
   if (diagnose && (strlen(temp_env) > 0)) {
      fprintf(stderr, "%s = %s\n", DEFAULT_LCC_ENV, temp_env);
   }
   temp_env[0]='\0';
   safecpy(temp_env, getenv(DEFAULT_OPT_ENV), MAX_LINELEN);
   if (diagnose && (strlen(temp_env) > 0)) {
      fprintf(stderr, "%s = %s\n", DEFAULT_OPT_ENV, temp_env);
   }
   
   if (result < 0) {
      if (diagnose) {
         fprintf(stderr, "%s exiting\n", argv[0]);
      }
      exit(result);
   }

   if (verbose > 1) {
      fprintf(stderr, "calling lcc\nlcc command = %s\n", lcc_cmd);
   }

   if ((result = system(lcc_cmd)) != 0) {
      if (verbose > 1) {
         fprintf(stderr, "lcc command returned result %d\n", result);
      }
   }

   if (diagnose) {
      fprintf(stderr, "\n%s done\n", argv[0]);
   }

   exit(result);
}
